from contextlib import closing
import hashlib
import os
import re
import shutil
import subprocess
import sys
import tarfile
from urllib import urlretrieve, urlopen
from urlparse import urlparse, urljoin
import zipfile


VERBOSE = False


class ALL(object):
    """
    Placeholder to select all resources, optional as well as required.
    """
    pass


class ResourceContainer(dict):
    def __init__(self, output_dir):
        super(ResourceContainer, self).__init__()
        self._required = set()
        self.output_dir = output_dir

    def add_required(self, name, resource):
        self[name] = Resource.get(name, resource, self.output_dir)
        self._required.add(name)

    def add_optional(self, name, resource):
        self[name] = Resource.get(name, resource, self.output_dir)

    def all(self):
        return self.values()

    def required(self):
        return [self[name] for name in self._required]

    def subset(self, which):
        if not which:
            return self.required()
        if which is ALL:
            return self.all()
        if not isinstance(which, list):
            return [self[which]]
        return [self[name] for name in which]


class Resource(object):
    """
    Base class for a Resource.

    Handles local file resources (with explicit ``filename`` or ``destination``).
    """
    @classmethod
    def get(cls, name, definition, output_dir):
        """
        Dispatch to the right subclass based on the definition.
        """
        if 'url' in definition:
            return URLResource(name, definition, output_dir)
        elif 'pypi' in definition:
            return PyPIResource(name, definition, output_dir)
        else:
            return Resource(name, definition, output_dir)

    def __init__(self, name, definition, output_dir):
        self.name = name
        self.source = definition.get('file', '')
        self.filename = os.path.basename(self.source)
        self.destination = definition.get(
            'destination', os.path.join(output_dir, self.filename))
        self.spec = self.destination
        self.hash = definition.get('hash', '')
        self.hash_type = definition.get('hash_type', '')
        self.output_dir = output_dir

    def fetch(self, mirror_url=None):
        return

    def verify(self):
        if self.hash_type not in hashlib.algorithms:
            return False
        if not os.path.isfile(self.destination):
            return False
        with open(self.destination) as fp:
            hash = hashlib.new(self.hash_type)
            hash.update(fp.read())
            if self.hash != hash.hexdigest():
                return False
        return True

    def install(self, destination, skip_top_level=False):
        if not self.verify():
            return False

        def filter_members(af):
            members = af.infolist() if hasattr(af, 'infolist') else af
            for member in members:
                if not skip_top_level:
                    yield member
                    continue
                if hasattr(member, 'path'):
                    path = member.path  # tarfiles
                elif hasattr(member, 'filename'):
                    path = member.filename  # zipfiles
                if re.match(r'^[^/]+/?$', path):
                    continue  # skip top-level members
                path = re.sub(r'^[^/]+/', '', path)  # strip top-level container
                if hasattr(member, 'path'):
                    member.path = path  # tarfiles
                elif hasattr(member, 'filename'):
                    member.filename = path  # zipfiles
                yield member

        if not os.path.exists(destination):
            os.makedirs(destination)

        if tarfile.is_tarfile(self.destination):
            with tarfile.open(self.destination) as tf:
                tf.extractall(destination, members=filter_members(tf))
        elif zipfile.is_zipfile(self.destination):
            with zipfile.ZipFile(self.destination, 'r') as zf:
                zf.extractall(destination, members=filter_members(zf))
        else:
            shutil.copy2(self.destination, destination)
        return True


class URLResource(Resource):
    def __init__(self, name, definition, output_dir):
        super(URLResource, self).__init__(name, definition, output_dir)
        self.url = definition.get('url', '')
        self.spec = self.url
        self.filename = definition.get(
            'filename', os.path.basename(urlparse(self.url).path))
        self.destination = definition.get(
            'destination', os.path.join(self.output_dir, self.filename))

    def fetch(self, mirror_url=None):
        if mirror_url:
            url = urljoin(mirror_url, self.filename)
        else:
            url = self.url
        if url.startswith('./'):
            url = url[2:]  # urlretrieve complains about this for some reason
        if not os.path.exists(os.path.dirname(self.destination)):
            os.makedirs(os.path.dirname(self.destination))
        try:
            urlretrieve(url, self.destination)
        except IOError as e:
            if VERBOSE:
                sys.stderr.write('Error fetching {}: {}\n'.format(self.url, e))
            # ignore download errors; they will be caught by verify


class PyPIResource(Resource):
    def __init__(self, name, definition, output_dir):
        super(PyPIResource, self).__init__(name, definition, output_dir)
        self.spec = definition.get('pypi', '')
        self.package_name = re.sub(r'[<>=].*', '', self.spec)
        self.destination_dir = os.path.join(self.output_dir, self.package_name)
        self.filename = ''
        self.destination = ''

    def fetch(self, mirror_url=None):
        if not os.path.exists(self.destination_dir):
            os.makedirs(self.destination_dir)
        cmd = ['pip', 'install', self.spec, '--download', self.destination_dir]
        if mirror_url:
            cmd.extend(['-i', mirror_url])
        try:
            subprocess.check_output(cmd, stderr=subprocess.STDOUT)
        except subprocess.CalledProcessError as e:  # noqa
            if VERBOSE:
                sys.stderr.write('Error fetching {}:\n{}\n'.format(self.name, e.output))
            return
        if not mirror_url:
            mirror_url = 'https://pypi.python.org/simple'
        mirror_url = mirror_url.rstrip('/') + '/'  # ensure trailing slash
        for filename in os.listdir(self.destination_dir):
            if filename.startswith(self.package_name):
                self.filename = filename
                self.destination = os.path.join(self.destination_dir, filename)
                hash_type, hash = self.get_remote_hash(self.filename, mirror_url)
                self.hash = hash
                self.hash_type = hash_type
                if hash_type:
                    hash_file = '.'.join([self.destination, self.hash_type])
                    self._write_file(hash_file, self.hash + '\n')
            else:
                self.process_dependency(filename, mirror_url)

    def verify(self):
        self.get_local_hash()
        return super(PyPIResource, self).verify()

    def get_local_hash(self):
        if not os.path.isdir(self.destination_dir):
            return
        for filename in os.listdir(self.destination_dir):
            fullname = os.path.join(self.destination_dir, filename)
            for hash_type in hashlib.algorithms:
                hash_file = '{}.{}'.format(fullname, hash_type)
                if os.path.isfile(hash_file):
                    self.filename = filename
                    self.destination = fullname
                    self.hash_type = hash_type
                    with open(hash_file) as fp:
                        self.hash = fp.readline().strip()
                    return

    def get_remote_hash(self, filename, mirror_url):
        package_name = self._package_name_from_filename(filename, mirror_url)
        url = urljoin(mirror_url, package_name)
        try:
            with closing(urlopen(url)) as fp:
                html = fp.read()
        except IOError as e:
            if VERBOSE:
                sys.stderr.write('Error fetching hash {}: {}\n'.format(url, e))
            return ('', '')
        link_re = (
            r'href=(?:"(?:[^"]*/)?|\'(?:[^\']*/)?)'
            '{}#([^=]+)=(\w+)["\']'.format(re.escape(filename)))
        match = re.search(link_re, html)
        if not match:
            if VERBOSE:
                sys.stderr.write('Hash not found for {}\n'.format(filename))
            return ('', '')
        return match.groups()

    def process_dependency(self, filename, mirror_url):
        # pip will download all dependencies into the same directory
        # we need to move them to their own package folders to be
        # properly mirrored
        package_name = self._package_name_from_filename(filename, mirror_url)
        new_dir = os.path.join(self.output_dir, package_name)
        old_dest = os.path.join(self.destination_dir, filename)
        new_dest = os.path.join(new_dir, filename)
        if not os.path.exists(new_dir):
            os.makedirs(new_dir)
        os.rename(old_dest, new_dest)
        hash_type, hash = self.get_remote_hash(filename, mirror_url)
        if hash_type:
            hash_file = '.'.join([new_dest, hash_type])
            self._write_file(hash_file, hash + '\n')

    @classmethod
    def _package_name_from_filename(cls, filename, mirror_url):
        # package file names may or may not contain various bits,
        # such as the arch, python version, package version, etc,
        # so we need to figure out what prefix is actually the
        # package name
        index = cls._get_index(mirror_url)
        parts = filename.split('-')
        while parts:
            package_name = '-'.join(parts)
            if package_name in index:
                return package_name
            parts.pop()
        return ''

    @classmethod
    def _get_index(cls, url):
        if not getattr(cls, '_index', None):
            try:
                with closing(urlopen(url)) as fp:
                    html = fp.read()
            except IOError as e:
                if VERBOSE:
                    sys.stderr.write('Error fetching index {}: {}\n'.format(url, e))
                html = ''
            cls._index = set(re.findall(r'<a href=(?:"[^"]*"|\'[^\']*\')>([^</]+)', html))
        return cls._index

    def _write_file(self, filename, text):
        with open(filename, 'w') as fp:
            fp.write(text)

    @classmethod
    def build_pypi_indexes(cls, root_dir):
        for entry in os.listdir(root_dir):
            candidate = os.path.join(root_dir, entry)
            if not os.path.isdir(candidate):
                continue
            res = PyPIResource(entry, {'pypi': entry}, root_dir)
            res.get_local_hash()
            if not res.hash_type:
                continue
            res._write_file(os.path.join(candidate, 'index.html'), '\n'.join([
                '<html>',
                '  <head>',
                '    <title>Links for {}</title>'.format(res.package_name),
                '    <meta name="api-version" value="2" />',
                '  </head>',
                '  <body>',
                '    <h1>Links for {}</h1>'.format(res.package_name),
                '    <a href="{0.filename}#{0.hash_type}={0.hash}" rel="internal">'
                '{0.filename}</h1>'.format(res),
                '  </body>',
                '</html>',
            ]))

    def install(self):
        if not self.verify():
            return False
        return subprocess.call(['pip', 'install', self.destination]) == 0

    @classmethod
    def install_group(cls, resources, mirror_url=None):
        to_install = []
        for resource in resources:
            if resource.verify():
                # use pre-fetched copy, if available
                to_install.append(resource.destination)
            else:
                # otherwise, try installing directly from mirror
                to_install.append(resource.spec)
        cmd = ['pip', 'install'] + to_install
        if mirror_url:
            cmd.extend(['-i', mirror_url])
        return subprocess.call(cmd) == 0



import contextlib
import subprocess
from urllib import urlopen

import yaml

from jujuresources.backend import ResourceContainer
from jujuresources.backend import PyPIResource
from jujuresources.backend import ALL


__all__ = ['fetch', 'verify', 'install', 'resource_path', 'resource_spec',
           'ALL', 'config_get', 'juju_log']
resources_cache = {}


def config_get(option_name):
    """
    Helper to access a Juju config option when charmhelpers is not available.

    :param str option_name: Name of the config option to get the value of
    """
    try:
        raw = subprocess.check_output(['config-get', option_name, '--format=yaml'])
        return yaml.load(raw.decode('UTF-8'))
    except ValueError:
        return None


def juju_log(message, level='DEBUG'):
    """
    Helper to send Juju log messages when charmhelpers is not available.

    :param str message: Message to log
    :param str level: Log level (DEBUG, INFO, ERROR, WARNING, CRITICAL; default: DEBUG)
    """
    subprocess.check_call(['juju-log', '-l', level, message])


def _load(resources_yaml, output_dir=None):
    if (resources_yaml, output_dir) not in resources_cache:
        with contextlib.closing(urlopen(resources_yaml)) as fp:
            resdefs = yaml.load(fp)
        _output_dir = output_dir or resdefs.get('options', {}).get('output_dir', 'resources')
        resources = ResourceContainer(_output_dir)
        for name, resource in resdefs.get('resources', {}).iteritems():
            resources.add_required(name, resource)
        for name, resource in resdefs.get('optional_resources', {}).iteritems():
            resources.add_optional(name, resource)
        resources_cache[(resources_yaml, output_dir)] = resources
    return resources_cache[(resources_yaml, output_dir)]


def _invalid(resources, which):
    invalid = set()
    for resource in resources.subset(which):
        if not resource.verify():
            invalid.add(resource.name)
    return invalid


def _fetch(resources, which, mirror_url, force=False, reporthook=None):
    invalid = _invalid(resources, which)
    for resource in resources.subset(which):
        if resource.name not in invalid and not force:
            continue
        if reporthook:
            reporthook(resource.name)
        resource.fetch(mirror_url)


def _install(resources, which, mirror_url, destination, skip_top_level):
    success = True
    pypi_resources = []
    for resource in resources.subset(which):
        if isinstance(resource, PyPIResource):
            # group pypi resources to reduce subprocess calls
            pypi_resources.append(resource)
        elif destination is not None:
            success = resource.install(destination, skip_top_level) and success
    if pypi_resources:
        success = PyPIResource.install_group(pypi_resources, mirror_url) and success
    return success


def invalid(which=None, resources_yaml='resources.yaml'):
    """
    Return a list of the names of the resources which do not
    pass :func:`verify`.

    :param list which: A name, or a list of one or more resource names, to
        fetch.  If ommitted, all non-optional resources are verified.
        You can also pass ``jujuresources.ALL`` to fetch all optional *and*
        required resources.
    :param str resources_yaml: Location of the yaml file containing the
        resource descriptions (default: ``./resources.yaml``).
        Can be a local file name or a remote URL.
    """
    resources = _load(resources_yaml, None)
    return _invalid(resources, which)


def verify(which=None, resources_yaml='resources.yaml'):
    """
    Verify if some or all resources previously fetched with :func:`fetch_resources`,
    including validating their cryptographic hash.

    :param list which: A list of one or more resource names to
        check.  If ommitted, all non-optional resources are verified.
        You can also pass ``jujuresources.ALL`` to fetch all optional and
        required resources.
    :param str resources_yaml: Location of the yaml file containing the
        resource descriptions (default: ``resources.yaml``).
        Can be a local file name or a remote URL.
    :param str output_dir: Override ``output_dir`` option from `resources_yaml`
        (this is intended for mirroring via the CLI and it is not recommended
        to be used otherwise)
    :return: True if all of the resources are available and valid, otherwise False.
    """
    resources = _load(resources_yaml, None)
    return not _invalid(resources, which)


def fetch(which=None, mirror_url=None, resources_yaml='resources.yaml',
          force=False, reporthook=None):
    """
    Attempt to fetch all resources for a charm.

    :param list which: A name, or a list of one or more resource names, to
        fetch.  If ommitted, all non-optional resources are fetched.
        You can also pass ``jujuresources.ALL`` to fetch all optional *and*
        required resources.
    :param str mirror_url: Fetch resources from the given mirror.
    :param str resources_yaml: Location of the yaml file containing the
        resource descriptions (default: ``./resources.yaml``).
        Can be a local file name or a remote URL.
    :param force bool: Force re-downloading of valid resources.
    :param func reporthook: Callback for reporting download progress.
        Will be called once for each resource, just prior to fetching, and will
        be passed the resource name.
    :return: True or False indicating whether the resources were successfully
        downloaded.
    """
    resources = _load(resources_yaml, None)
    _fetch(resources, which, mirror_url, force, reporthook)
    return not _invalid(resources, which)


def resource_path(resource_name, resources_yaml='resources.yaml'):
    """
    Get the local path for a named resource that has been fetched.

    This may return ``None`` if the local path cannot be determined
    (for example, if the resource has not been fetched yet and needs
    to be resolved).  Even if it returns a path, that path is not
    guaranteed to exist or be valid; you should always confirm that
    a resource is available using :func:`verify` or :func:`fetch`
    before using it.

    :param str resource_name: The name of a resource to resolve.
    :param str resources_yaml: Location of the yaml file containing the
        resource descriptions (default: ``./resources.yaml``).
        Can be a local file name or a remote URL.
    """
    resources = _load(resources_yaml, None)
    return resources[resource_name].destination


def resource_spec(resource_name, resources_yaml='resources.yaml'):
    """
    Get the spec for a named resource.  This would be the URL for URL
    resources, the Python package spec for PyPI resources, the full
    path for local file resources, etc.

    :param str resource_name: The name of a resource to resolve.
    :param str resources_yaml: Location of the yaml file containing the
        resource descriptions (default: ``./resources.yaml``).
        Can be a local file name or a remote URL.
    """
    resources = _load(resources_yaml, None)
    return resources[resource_name].spec


def install(which=None, mirror_url=None, destination=None, skip_top_level=False,
            resources_yaml='resources.yaml'):
    """
    Install one or more resources.

    The resource(s) will be fetched, if necessary, and different resource
    types are handled appropriately (e.g., PyPI resources are installed
    with ``pip``, archive file resources are extracted, non-archive file
    resources are copied, etc).

    For PyPI resources, this is roughly equivalent to the following::

        pip install `juju-resources resource_spec $resource` -i $mirror_url

    :param list which: A name, or a list of one or more resource names, to
        fetch.  If ommitted, all non-optional resources are installed.
    :param str mirror_url: Fetch resources from the given mirror.
    :param str destination: Destination to which to extract or copy file resources.
    :param bool skip_top_level: When extracting archive file resources, skip
        all members that are at the top level of the archive and instead extract
        all nested members directly into ``destination``.  E.g., an archive
        containing ``foo/bar.txt`` and ``foo/qux/baz.txt`` will be extracted as
        ``destination/bar.txt`` and ``destination/qux/baz.txt``.
    :param str resources_yaml: Location of the yaml file containing the
        resource descriptions (default: ``resources.yaml``).
        Can be a local file name or a remote URL.
    :returns: True if all resources were successfully installed.
    """
    resources = _load(resources_yaml, None)
    return _install(resources, mirror_url, destination, skip_top_level)



from setuptools import setup
import os


version_file = os.path.abspath(os.path.join(os.path.dirname(__file__),
                                            'VERSION'))
with open(version_file) as v:
    VERSION = v.read().strip()


SETUP = {
    'name': "jujuresources",
    'version': VERSION,
    'author': "Cory Johns",
    'author_email': "cory.johns@canonical.com",
    'url': "https://github.com/juju-solutions/jujuresources",
    'packages': [
        "jujuresources",
    ],
    'install_requires': [
        'pyaml',
    ],
    'entry_points': {
        'console_scripts': [
            'juju-resources = jujuresources.cli:resources',
        ],
        'jujuresources.subcommands': [
            'fetch = jujuresources.cli:fetch',
            'verify = jujuresources.cli:verify',
            'serve = jujuresources.cli:serve',
            'resource_path = jujuresources.cli:resource_path',
        ],
    },
    'license': "MIT License",
    'long_description': open('README.rst').read(),
    'description': 'Helpers for Juju Charms to load external resources',
}

try:
    from sphinx_pypi_upload import UploadDoc
    SETUP['cmdclass'] = {'upload_sphinx': UploadDoc}
except ImportError:
    pass

if __name__ == '__main__':
    setup(**SETUP)



from __future__ import print_function
import os
import ssl
import sys
import socket
import argparse
from pkg_resources import iter_entry_points
from SimpleHTTPServer import SimpleHTTPRequestHandler
import SocketServer

from jujuresources import _fetch
from jujuresources import _install
from jujuresources import _invalid
from jujuresources import _load
from jujuresources import ALL
from jujuresources import backend


def arg(*args, **kwargs):
    """
    Decorator to add args to subcommands.
    """
    def _arg(f):
        if not hasattr(f, '_subcommand_args'):
            f._subcommand_args = []
        f._subcommand_args.append((args, kwargs))
        return f
    return _arg


def argset(name, *args, **kwargs):
    """
    Decorator to add sets of required mutually exclusive args to subcommands.
    """
    def _arg(f):
        if not hasattr(f, '_subcommand_argsets'):
            f._subcommand_argsets = {}
        f._subcommand_argsets.setdefault(name, []).append((args, kwargs))
        return f
    return _arg


print = print  # for testing
_exit = sys.exit  # for testing


def resources(argv=sys.argv[1:]):
    """
    Juju CLI subcommand for dispatching resources subcommands.
    """
    eps = iter_entry_points('jujuresources.subcommands')
    ep_map = {ep.name: ep.load() for ep in eps}

    parser = argparse.ArgumentParser()
    if '--description' in argv:
        print('Manage and mirror charm resources')
        return 0

    subparsers = {}
    subparser_factory = parser.add_subparsers()
    subparsers['help'] = subparser_factory.add_parser('help', help='Display help for a subcommand')
    subparsers['help'].add_argument('command', nargs='?')
    subparsers['help'].set_defaults(subcommand='help')
    for name, subcommand in ep_map.iteritems():
        subparsers[name] = subparser_factory.add_parser(name, help=subcommand.__doc__)
        subparsers[name].set_defaults(subcommand=subcommand)
        for args, kwargs in getattr(subcommand, '_subcommand_args', []):
            subparsers[name].add_argument(*args, **kwargs)
        for argset in getattr(subcommand, '_subcommand_argsets', {}).values():
            group = subparsers[name].add_mutually_exclusive_group(required=True)
            for args, kwargs in argset:
                group.add_argument(*args, **kwargs)
    opts = parser.parse_args(argv)
    if opts.subcommand == 'help':
        if opts.command:
            subparsers[opts.command].print_help()
        else:
            parser.print_help()
    else:
        return _exit(opts.subcommand(opts) or 0)


@arg('-r', '--resources', default='resources.yaml',
     help='File or URL containing the YAML resource descriptions (default: ./resources.yaml)')
@arg('-d', '--output-dir', default=None,
     help='Directory to place the fetched resources (default: ./resources/)')
@arg('-u', '--mirror-url',
     help='URL at which the resources are mirrored')
@arg('-a', '--all', action='store_true',
     help='Include all optional resources as well as required')
@arg('-q', '--quiet', action='store_true',
     help='Suppress output and only set the return code')
@arg('-f', '--force', action='store_true',
     help='Force re-download of valid resources')
@arg('-v', '--verbose', action='store_true',
     help='Write download error information to stderr')
@arg('resource_names', nargs='*',
     help='Names of specific resources to fetch (defaults to all required, '
          'or all if --all is given)')
def fetch(opts):
    """
    Create a local mirror of one or more resources.
    """
    resources = _load(opts.resources, opts.output_dir)
    if opts.all:
        opts.resource_names = ALL
    reporthook = None if opts.quiet else lambda name: print('Fetching {}...'.format(name))
    if opts.verbose:
        backend.VERBOSE = True
    _fetch(resources, opts.resource_names, opts.mirror_url, opts.force, reporthook)
    return verify(opts)


@arg('-r', '--resources', default='resources.yaml',
     help='File or URL containing the YAML resource descriptions (default: ./resources.yaml)')
@arg('-d', '--output-dir', default=None,
     help='Directory containing the fetched resources (default: ./resources/)')
@arg('-a', '--all', action='store_true',
     help='Include all optional resources as well as required')
@arg('-q', '--quiet', action='store_true',
     help='Suppress output and only set the return code')
@arg('resource_names', nargs='*',
     help='Names of specific resources to verify (defaults to all required, '
          'or all if --all is given)')
def verify(opts):
    """
    Verify that one or more resources were downloaded successfully.
    """
    resources = _load(opts.resources, opts.output_dir)
    if opts.all:
        opts.resource_names = ALL
    invalid = _invalid(resources, opts.resource_names)
    if not invalid:
        if not opts.quiet:
            print("All resources successfully downloaded")
        return 0
    else:
        if not opts.quiet:
            print("Invalid or missing resources: {}".format(', '.join(invalid)))
        return 1


@arg('-r', '--resources', default='resources.yaml',
     help='File or URL containing the YAML resource descriptions (default: ./resources.yaml)')
@arg('-d', '--output-dir', default=None,
     help='Directory containing the fetched resources (default: ./resources/)')
@arg('-u', '--mirror-url',
     help='URL at which the resources are mirrored')
@arg('-a', '--all', action='store_true',
     help='Include all optional resources as well as required')
@arg('-q', '--quiet', action='store_true',
     help='Suppress output and only set the return code')
@arg('-D', '--destination',
     help='Destination for archive or file resources to be installed to')
@arg('-s', '--skip-top-level', action='store_true',
     help='Skip top-level members of archives, and extract children directly to destination')
@arg('resource_names', nargs='*',
     help='Names of specific resources to verify (defaults to all required, '
          'or all if --all is given)')
def install(opts):
    """
    Install one or more resources.
    """
    resources = _load(opts.resources, opts.output_dir)
    if opts.all:
        opts.resource_names = ALL
    success = _install(resources, opts.resource_names, opts.mirror_url,
                       opts.destination, opts.skip_top_level)
    if success:
        if not opts.quiet:
            print("All resources successfully installed")
        return 0
    else:
        if not opts.quiet:
            invalid = _invalid(resources, opts.resource_names)
            print("Unable to install some resources: {}".format(', '.join(invalid)))
        return 1


@arg('-r', '--resources', default='resources.yaml',
     help='File or URL containing the YAML resource descriptions (default: ./resources.yaml)')
@arg('-d', '--output-dir', default=None,
     help='Directory containing the fetched resources (default: ./resources/)')
@arg('resource_name', help='Name of a resource')
def resource_path(opts):
    """
    Return the full path to a named resource.
    """
    resources = _load(opts.resources, opts.output_dir)
    if opts.resource_name not in resources:
        sys.stderr.write('Invalid resource name: {}\n'.format(opts.resource_name))
        return 1
    print(resources[opts.resource_name].destination)


@arg('-r', '--resources', default='resources.yaml',
     help='File or URL containing the YAML resource descriptions (default: ./resources.yaml)')
@arg('-d', '--output-dir', default=None,
     help='Directory containing the fetched resources (default: ./resources/)')
@arg('resource_name', help='Name of a resource')
def resource_spec(opts):
    """
    Return the spec (URL, package spec, file, etc) for a named resource.
    """
    resources = _load(opts.resources, opts.output_dir)
    if opts.resource_name not in resources:
        sys.stderr.write('Invalid resource name: {}\n'.format(opts.resource_name))
        return 1
    print(resources[opts.resource_name].spec)


@arg('-r', '--resources', default='resources.yaml',
     help='File or URL containing the YAML resource descriptions (default: ./resources.yaml)')
@arg('-d', '--output-dir', default=None,
     help='Directory containing the fetched resources (default: ./resources/)')
@arg('-H', '--host', default='',
     help='IP address on which to bind the mirror server')
@arg('-p', '--port', type=int, default=8080,
     help='Port on which to bind the mirror server')
@arg('-s', '--ssl-cert', default=None,
     help='Path to an SSL certificate file (will run without SSL if not given)')
def serve(opts):
    """
    Run a light-weight HTTP server hosting previously mirrored resources
    """
    resources = _load(opts.resources, opts.output_dir)
    opts.output_dir = resources.output_dir  # allow resources.yaml to set default output_dir
    if not os.path.exists(opts.output_dir):
        sys.stderr.write("Resources dir '{}' not found.  Did you fetch?\n".format(opts.output_dir))
        return 1
    backend.PyPIResource.build_pypi_indexes(opts.output_dir)
    os.chdir(opts.output_dir)

    SocketServer.TCPServer.allow_reuse_address = True
    httpd = SocketServer.TCPServer((opts.host, opts.port), SimpleHTTPRequestHandler)

    if opts.ssl_cert:
        httpd.socket = ssl.wrap_socket(httpd.socket, certfile=opts.ssl_cert, server_side=True)

    print("Serving at: http{}://{}:{}/".format(
        's' if opts.ssl_cert else '', socket.gethostname(), opts.port))
    httpd.serve_forever()
